/* IPFIX entities. */
#ifndef IPFIX_ENTITY
#define IPFIX_ENTITY(ENUM, ID, SIZE, NAME)
#endif

IPFIX_ENTITY(OCTET_DELTA_COUNT, 1, 8, octetDeltaCount)
IPFIX_ENTITY(PACKET_DELTA_COUNT, 2, 8, packetDeltaCount)
IPFIX_ENTITY(DELTA_FLOW_COUNT, 3, 8, deltaFlowCount)
IPFIX_ENTITY(PROTOCOL_IDENTIFIER, 4, 1, protocolIdentifier)
IPFIX_ENTITY(IP_CLASS_OF_SERVICE, 5, 1, ipClassOfService)
IPFIX_ENTITY(TCP_CONTROL_BITS, 6, 1, tcpControlBits)
IPFIX_ENTITY(SOURCE_TRANSPORT_PORT, 7, 2, sourceTransportPort)
IPFIX_ENTITY(SOURCE_IPV4_ADDRESS, 8, 4, sourceIPv4Address)
IPFIX_ENTITY(SOURCE_IPV4_PREFIX_LENGTH, 9, 1, sourceIPv4PrefixLength)
IPFIX_ENTITY(INGRESS_INTERFACE, 10, 4, ingressInterface)
IPFIX_ENTITY(DESTINATION_TRANSPORT_PORT, 11, 2, destinationTransportPort)
IPFIX_ENTITY(DESTINATION_IPV4_ADDRESS, 12, 4, destinationIPv4Address)
IPFIX_ENTITY(DESTINATION_IPV4_PREFIX_LENGTH, 13, 1, destinationIPv4PrefixLength)
IPFIX_ENTITY(EGRESS_INTERFACE, 14, 4, egressInterface)
IPFIX_ENTITY(IP_NEXT_HOP_IPV4_ADDRESS, 15, 4, ipNextHopIPv4Address)
IPFIX_ENTITY(BGP_SOURCE_AS_NUMBER, 16, 4, bgpSourceAsNumber)
IPFIX_ENTITY(BGP_DESTINATION_AS_NUMBER, 17, 4, bgpDestinationAsNumber)
IPFIX_ENTITY(BGP_NEXT_HOP_IPV4_ADDRESS, 18, 4, bgpNextHopIPv4Address)
IPFIX_ENTITY(POST_MCAST_PACKET_DELTA_COUNT, 19, 8, postMCastPacketDeltaCount)
IPFIX_ENTITY(POST_MCAST_OCTET_DELTA_COUNT, 20, 8, postMCastOctetDeltaCount)
IPFIX_ENTITY(FLOW_END_SYS_UP_TIME, 21, 4, flowEndSysUpTime)
IPFIX_ENTITY(FLOW_START_SYS_UP_TIME, 22, 4, flowStartSysUpTime)
IPFIX_ENTITY(POST_OCTET_DELTA_COUNT, 23, 8, postOctetDeltaCount)
IPFIX_ENTITY(POST_PACKET_DELTA_COUNT, 24, 8, postPacketDeltaCount)
IPFIX_ENTITY(MINIMUM_IP_TOTAL_LENGTH, 25, 8, minimumIpTotalLength)
IPFIX_ENTITY(MAXIMUM_IP_TOTAL_LENGTH, 26, 8, maximumIpTotalLength)
IPFIX_ENTITY(SOURCE_IPV6_ADDRESS, 27, 16, sourceIPv6Address)
IPFIX_ENTITY(DESTINATION_IPV6_ADDRESS, 28, 16, destinationIPv6Address)
IPFIX_ENTITY(SOURCE_IPV6_PREFIX_LENGTH, 29, 1, sourceIPv6PrefixLength)
IPFIX_ENTITY(DESTINATION_IPV6_PREFIX_LENGTH, 30, 1, destinationIPv6PrefixLength)
IPFIX_ENTITY(FLOW_LABEL_IPV6, 31, 4, flowLabelIPv6)
IPFIX_ENTITY(ICMP_TYPE_CODE_IPV4, 32, 2, icmpTypeCodeIPv4)
IPFIX_ENTITY(IGMP_TYPE, 33, 1, igmpType)
IPFIX_ENTITY(FLOW_ACTIVE_TIMEOUT, 36, 2, flowActiveTimeout)
IPFIX_ENTITY(FLOW_IDLE_TIMEOUT, 37, 2, flowIdleTimeout)
IPFIX_ENTITY(EXPORTED_OCTET_TOTAL_COUNT, 40, 8, exportedOctetTotalCount)
IPFIX_ENTITY(EXPORTED_MESSAGE_TOTAL_COUNT, 41, 8, exportedMessageTotalCount)
IPFIX_ENTITY(EXPORTED_FLOW_RECORD_TOTAL_COUNT, 42, 8, exportedFlowRecordTotalCount)
IPFIX_ENTITY(SOURCE_IPV4_PREFIX, 44, 4, sourceIPv4Prefix)
IPFIX_ENTITY(DESTINATION_IPV4_PREFIX, 45, 4, destinationIPv4Prefix)
IPFIX_ENTITY(MPLS_TOP_LABEL_TYPE, 46, 1, mplsTopLabelType)
IPFIX_ENTITY(MPLS_TOP_LABEL_IPV4_ADDRESS, 47, 4, mplsTopLabelIPv4Address)
IPFIX_ENTITY(MINIMUM_TTL, 52, 1, minimumTTL)
IPFIX_ENTITY(MAXIMUM_TTL, 53, 1, maximumTTL)
IPFIX_ENTITY(FRAGMENT_IDENTIFICATION, 54, 4, fragmentIdentification)
IPFIX_ENTITY(POST_IP_CLASS_OF_SERVICE, 55, 1, postIpClassOfService)
IPFIX_ENTITY(SOURCE_MAC_ADDRESS, 56, 6, sourceMacAddress)
IPFIX_ENTITY(POST_DESTINATION_MAC_ADDRESS, 57, 6, postDestinationMacAddress)
IPFIX_ENTITY(VLAN_ID, 58, 2, vlanId)
IPFIX_ENTITY(POST_VLAN_ID, 59, 2, postVlanId)
IPFIX_ENTITY(IP_VERSION, 60, 1, ipVersion)
IPFIX_ENTITY(FLOW_DIRECTION, 61, 1, flowDirection)
IPFIX_ENTITY(IP_NEXT_HOP_IPV6_ADDRESS, 62, 16, ipNextHopIPv6Address)
IPFIX_ENTITY(BGP_NEXT_HOP_IPV6_ADDRESS, 63, 16, bgpNextHopIPv6Address)
IPFIX_ENTITY(IPV6_EXTENSION_HEADERS, 64, 4, ipv6ExtensionHeaders)
IPFIX_ENTITY(MPLS_TOP_LABEL_STACK_SECTION, 70, 0, mplsTopLabelStackSection)
IPFIX_ENTITY(MPLS_LABEL_STACK_SECTION2, 71, 0, mplsLabelStackSection2)
IPFIX_ENTITY(MPLS_LABEL_STACK_SECTION3, 72, 0, mplsLabelStackSection3)
IPFIX_ENTITY(MPLS_LABEL_STACK_SECTION4, 73, 0, mplsLabelStackSection4)
IPFIX_ENTITY(MPLS_LABEL_STACK_SECTION5, 74, 0, mplsLabelStackSection5)
IPFIX_ENTITY(MPLS_LABEL_STACK_SECTION6, 75, 0, mplsLabelStackSection6)
IPFIX_ENTITY(MPLS_LABEL_STACK_SECTION7, 76, 0, mplsLabelStackSection7)
IPFIX_ENTITY(MPLS_LABEL_STACK_SECTION8, 77, 0, mplsLabelStackSection8)
IPFIX_ENTITY(MPLS_LABEL_STACK_SECTION9, 78, 0, mplsLabelStackSection9)
IPFIX_ENTITY(MPLS_LABEL_STACK_SECTION10, 79, 0, mplsLabelStackSection10)
IPFIX_ENTITY(DESTINATION_MAC_ADDRESS, 80, 6, destinationMacAddress)
IPFIX_ENTITY(POST_SOURCE_MAC_ADDRESS, 81, 6, postSourceMacAddress)
IPFIX_ENTITY(INTERFACE_NAME, 82, 0, interfaceName)
IPFIX_ENTITY(INTERFACE_DESCRIPTION, 83, 0, interfaceDescription)
IPFIX_ENTITY(OCTET_TOTAL_COUNT, 85, 8, octetTotalCount)
IPFIX_ENTITY(PACKET_TOTAL_COUNT, 86, 8, packetTotalCount)
IPFIX_ENTITY(FRAGMENT_OFFSET, 88, 2, fragmentOffset)
IPFIX_ENTITY(MPLS_VPN_ROUTE_DISTINGUISHER, 90, 0, mplsVpnRouteDistinguisher)
IPFIX_ENTITY(MPLS_TOP_LABEL_PREFIX_LENGTH, 91, 1, mplsTopLabelPrefixLength)
IPFIX_ENTITY(APPLICATION_DESCRIPTION, 94, 0, applicationDescription)
IPFIX_ENTITY(APPLICATION_ID, 95, 0, applicationId)
IPFIX_ENTITY(APPLICATION_NAME, 96, 0, applicationName)
IPFIX_ENTITY(POST_IP_DIFF_SERV_CODE_POINT, 98, 1, postIpDiffServCodePoint)
IPFIX_ENTITY(MULTICAST_REPLICATION_FACTOR, 99, 4, multicastReplicationFactor)
IPFIX_ENTITY(CLASSIFICATION_ENGINE_ID, 101, 1, classificationEngineId)
IPFIX_ENTITY(BGP_NEXT_ADJACENT_AS_NUMBER, 128, 4, bgpNextAdjacentAsNumber)
IPFIX_ENTITY(BGP_PREV_ADJACENT_AS_NUMBER, 129, 4, bgpPrevAdjacentAsNumber)
IPFIX_ENTITY(EXPORTER_IPV4_ADDRESS, 130, 4, exporterIPv4Address)
IPFIX_ENTITY(EXPORTER_IPV6_ADDRESS, 131, 16, exporterIPv6Address)
IPFIX_ENTITY(DROPPED_OCTET_DELTA_COUNT, 132, 8, droppedOctetDeltaCount)
IPFIX_ENTITY(DROPPED_PACKET_DELTA_COUNT, 133, 8, droppedPacketDeltaCount)
IPFIX_ENTITY(DROPPED_OCTET_TOTAL_COUNT, 134, 8, droppedOctetTotalCount)
IPFIX_ENTITY(DROPPED_PACKET_TOTAL_COUNT, 135, 8, droppedPacketTotalCount)
IPFIX_ENTITY(FLOW_END_REASON, 136, 1, flowEndReason)
IPFIX_ENTITY(COMMON_PROPERTIES_ID, 137, 8, commonPropertiesId)
IPFIX_ENTITY(OBSERVATION_POINT_ID, 138, 4, observationPointId)
IPFIX_ENTITY(ICMP_TYPE_CODE_IPV6, 139, 2, icmpTypeCodeIPv6)
IPFIX_ENTITY(MPLS_TOP_LABEL_IPV6_ADDRESS, 140, 16, mplsTopLabelIPv6Address)
IPFIX_ENTITY(LINE_CARD_ID, 141, 4, lineCardId)
IPFIX_ENTITY(PORT_ID, 142, 4, portId)
IPFIX_ENTITY(METERING_PROCESS_ID, 143, 4, meteringProcessId)
IPFIX_ENTITY(EXPORTING_PROCESS_ID, 144, 4, exportingProcessId)
IPFIX_ENTITY(TEMPLATE_ID, 145, 2, templateId)
IPFIX_ENTITY(WLAN_CHANNEL_ID, 146, 1, wlanChannelId)
IPFIX_ENTITY(WLAN_SSID, 147, 0, wlanSSID)
IPFIX_ENTITY(FLOW_ID, 148, 8, flowId)
IPFIX_ENTITY(OBSERVATION_DOMAIN_ID, 149, 4, observationDomainId)
IPFIX_ENTITY(FLOW_START_SECONDS, 150, 4, flowStartSeconds)
IPFIX_ENTITY(FLOW_END_SECONDS, 151, 4, flowEndSeconds)
IPFIX_ENTITY(FLOW_START_MILLISECONDS, 152, 8, flowStartMilliseconds)
IPFIX_ENTITY(FLOW_END_MILLISECONDS, 153, 8, flowEndMilliseconds)
IPFIX_ENTITY(FLOW_START_MICROSECONDS, 154, 8, flowStartMicroseconds)
IPFIX_ENTITY(FLOW_END_MICROSECONDS, 155, 8, flowEndMicroseconds)
IPFIX_ENTITY(FLOW_START_NANOSECONDS, 156, 8, flowStartNanoseconds)
IPFIX_ENTITY(FLOW_END_NANOSECONDS, 157, 8, flowEndNanoseconds)
IPFIX_ENTITY(FLOW_START_DELTA_MICROSECONDS, 158, 4, flowStartDeltaMicroseconds)
IPFIX_ENTITY(FLOW_END_DELTA_MICROSECONDS, 159, 4, flowEndDeltaMicroseconds)
IPFIX_ENTITY(SYSTEM_INIT_TIME_MILLISECONDS, 160, 8, systemInitTimeMilliseconds)
IPFIX_ENTITY(FLOW_DURATION_MILLISECONDS, 161, 4, flowDurationMilliseconds)
IPFIX_ENTITY(FLOW_DURATION_MICROSECONDS, 162, 4, flowDurationMicroseconds)
IPFIX_ENTITY(OBSERVED_FLOW_TOTAL_COUNT, 163, 8, observedFlowTotalCount)
IPFIX_ENTITY(IGNORED_PACKET_TOTAL_COUNT, 164, 8, ignoredPacketTotalCount)
IPFIX_ENTITY(IGNORED_OCTET_TOTAL_COUNT, 165, 8, ignoredOctetTotalCount)
IPFIX_ENTITY(NOT_SENT_FLOW_TOTAL_COUNT, 166, 8, notSentFlowTotalCount)
IPFIX_ENTITY(NOT_SENT_PACKET_TOTAL_COUNT, 167, 8, notSentPacketTotalCount)
IPFIX_ENTITY(NOT_SENT_OCTET_TOTAL_COUNT, 168, 8, notSentOctetTotalCount)
IPFIX_ENTITY(DESTINATION_IPV6_PREFIX, 169, 16, destinationIPv6Prefix)
IPFIX_ENTITY(SOURCE_IPV6_PREFIX, 170, 16, sourceIPv6Prefix)
IPFIX_ENTITY(POST_OCTET_TOTAL_COUNT, 171, 8, postOctetTotalCount)
IPFIX_ENTITY(POST_PACKET_TOTAL_COUNT, 172, 8, postPacketTotalCount)
IPFIX_ENTITY(FLOW_KEY_INDICATOR, 173, 8, flowKeyIndicator)
IPFIX_ENTITY(POST_MCAST_PACKET_TOTAL_COUNT, 174, 8, postMCastPacketTotalCount)
IPFIX_ENTITY(POST_MCAST_OCTET_TOTAL_COUNT, 175, 8, postMCastOctetTotalCount)
IPFIX_ENTITY(ICMP_TYPE_IPV4, 176, 1, icmpTypeIPv4)
IPFIX_ENTITY(ICMP_CODE_IPV4, 177, 1, icmpCodeIPv4)
IPFIX_ENTITY(ICMP_TYPE_IPV6, 178, 1, icmpTypeIPv6)
IPFIX_ENTITY(ICMP_CODE_IPV6, 179, 1, icmpCodeIPv6)
IPFIX_ENTITY(UDP_SOURCE_PORT, 180, 2, udpSourcePort)
IPFIX_ENTITY(UDP_DESTINATION_PORT, 181, 2, udpDestinationPort)
IPFIX_ENTITY(TCP_SOURCE_PORT, 182, 2, tcpSourcePort)
IPFIX_ENTITY(TCP_DESTINATION_PORT, 183, 2, tcpDestinationPort)
IPFIX_ENTITY(TCP_SEQUENCE_NUMBER, 184, 4, tcpSequenceNumber)
IPFIX_ENTITY(TCP_ACKNOWLEDGEMENT_NUMBER, 185, 4, tcpAcknowledgementNumber)
IPFIX_ENTITY(TCP_WINDOW_SIZE, 186, 2, tcpWindowSize)
IPFIX_ENTITY(TCP_URGENT_POINTER, 187, 2, tcpUrgentPointer)
IPFIX_ENTITY(TCP_HEADER_LENGTH, 188, 1, tcpHeaderLength)
IPFIX_ENTITY(IP_HEADER_LENGTH, 189, 1, ipHeaderLength)
IPFIX_ENTITY(TOTAL_LENGTH_IPV4, 190, 2, totalLengthIPv4)
IPFIX_ENTITY(PAYLOAD_LENGTH_IPV6, 191, 2, payloadLengthIPv6)
IPFIX_ENTITY(IP_TTL, 192, 1, ipTTL)
IPFIX_ENTITY(NEXT_HEADER_IPV6, 193, 1, nextHeaderIPv6)
IPFIX_ENTITY(MPLS_PAYLOAD_LENGTH, 194, 4, mplsPayloadLength)
IPFIX_ENTITY(IP_DIFF_SERV_CODE_POINT, 195, 1, ipDiffServCodePoint)
IPFIX_ENTITY(IP_PRECEDENCE, 196, 1, ipPrecedence)
IPFIX_ENTITY(FRAGMENT_FLAGS, 197, 1, fragmentFlags)
IPFIX_ENTITY(OCTET_DELTA_SUM_OF_SQUARES, 198, 8, octetDeltaSumOfSquares)
IPFIX_ENTITY(OCTET_TOTAL_SUM_OF_SQUARES, 199, 8, octetTotalSumOfSquares)
IPFIX_ENTITY(MPLS_TOP_LABEL_TTL, 200, 1, mplsTopLabelTTL)
IPFIX_ENTITY(MPLS_LABEL_STACK_LENGTH, 201, 4, mplsLabelStackLength)
IPFIX_ENTITY(MPLS_LABEL_STACK_DEPTH, 202, 4, mplsLabelStackDepth)
IPFIX_ENTITY(MPLS_TOP_LABEL_EXP, 203, 1, mplsTopLabelExp)
IPFIX_ENTITY(IP_PAYLOAD_LENGTH, 204, 4, ipPayloadLength)
IPFIX_ENTITY(UDP_MESSAGE_LENGTH, 205, 2, udpMessageLength)
IPFIX_ENTITY(IS_MULTICAST, 206, 1, isMulticast)
IPFIX_ENTITY(IPV4_IHL, 207, 1, ipv4IHL)
IPFIX_ENTITY(IPV4_OPTIONS, 208, 4, ipv4Options)
IPFIX_ENTITY(TCP_OPTIONS, 209, 8, tcpOptions)
IPFIX_ENTITY(PADDING_OCTETS, 210, 0, paddingOctets)
IPFIX_ENTITY(COLLECTOR_IPV4_ADDRESS, 211, 4, collectorIPv4Address)
IPFIX_ENTITY(COLLECTOR_IPV6_ADDRESS, 212, 16, collectorIPv6Address)
IPFIX_ENTITY(EXPORT_INTERFACE, 213, 4, exportInterface)
IPFIX_ENTITY(EXPORT_PROTOCOL_VERSION, 214, 1, exportProtocolVersion)
IPFIX_ENTITY(EXPORT_TRANSPORT_PROTOCOL, 215, 1, exportTransportProtocol)
IPFIX_ENTITY(COLLECTOR_TRANSPORT_PORT, 216, 2, collectorTransportPort)
IPFIX_ENTITY(EXPORTER_TRANSPORT_PORT, 217, 2, exporterTransportPort)
IPFIX_ENTITY(TCP_SYN_TOTAL_COUNT, 218, 8, tcpSynTotalCount)
IPFIX_ENTITY(TCP_FIN_TOTAL_COUNT, 219, 8, tcpFinTotalCount)
IPFIX_ENTITY(TCP_RST_TOTAL_COUNT, 220, 8, tcpRstTotalCount)
IPFIX_ENTITY(TCP_PSH_TOTAL_COUNT, 221, 8, tcpPshTotalCount)
IPFIX_ENTITY(TCP_ACK_TOTAL_COUNT, 222, 8, tcpAckTotalCount)
IPFIX_ENTITY(TCP_URG_TOTAL_COUNT, 223, 8, tcpUrgTotalCount)
IPFIX_ENTITY(IP_TOTAL_LENGTH, 224, 8, ipTotalLength)
IPFIX_ENTITY(POST_NATSOURCE_IPV4_ADDRESS, 225, 4, postNATSourceIPv4Address)
IPFIX_ENTITY(POST_NATDESTINATION_IPV4_ADDRESS, 226, 4, postNATDestinationIPv4Address)
IPFIX_ENTITY(POST_NAPTSOURCE_TRANSPORT_PORT, 227, 2, postNAPTSourceTransportPort)
IPFIX_ENTITY(POST_NAPTDESTINATION_TRANSPORT_PORT, 228, 2, postNAPTDestinationTransportPort)
IPFIX_ENTITY(NAT_ORIGINATING_ADDRESS_REALM, 229, 1, natOriginatingAddressRealm)
IPFIX_ENTITY(NAT_EVENT, 230, 1, natEvent)
IPFIX_ENTITY(INITIATOR_OCTETS, 231, 8, initiatorOctets)
IPFIX_ENTITY(RESPONDER_OCTETS, 232, 8, responderOctets)
IPFIX_ENTITY(FIREWALL_EVENT, 233, 1, firewallEvent)
IPFIX_ENTITY(INGRESS_VRFID, 234, 4, ingressVRFID)
IPFIX_ENTITY(EGRESS_VRFID, 235, 4, egressVRFID)
IPFIX_ENTITY(V_RFNAME, 236, 0, VRFname)
IPFIX_ENTITY(POST_MPLS_TOP_LABEL_EXP, 237, 1, postMplsTopLabelExp)
IPFIX_ENTITY(TCP_WINDOW_SCALE, 238, 2, tcpWindowScale)
IPFIX_ENTITY(BIFLOW_DIRECTION, 239, 1, biflowDirection)
IPFIX_ENTITY(ETHERNET_HEADER_LENGTH, 240, 1, ethernetHeaderLength)
IPFIX_ENTITY(ETHERNET_PAYLOAD_LENGTH, 241, 2, ethernetPayloadLength)
IPFIX_ENTITY(ETHERNET_TOTAL_LENGTH, 242, 2, ethernetTotalLength)
IPFIX_ENTITY(DOT1Q_VLAN_ID, 243, 2, dot1qVlanId)
IPFIX_ENTITY(DOT1Q_PRIORITY, 244, 1, dot1qPriority)
IPFIX_ENTITY(DOT1Q_CUSTOMER_VLAN_ID, 245, 2, dot1qCustomerVlanId)
IPFIX_ENTITY(DOT1Q_CUSTOMER_PRIORITY, 246, 1, dot1qCustomerPriority)
IPFIX_ENTITY(METRO_EVC_ID, 247, 0, metroEvcId)
IPFIX_ENTITY(METRO_EVC_TYPE, 248, 1, metroEvcType)
IPFIX_ENTITY(PSEUDO_WIRE_ID, 249, 4, pseudoWireId)
IPFIX_ENTITY(PSEUDO_WIRE_TYPE, 250, 2, pseudoWireType)
IPFIX_ENTITY(PSEUDO_WIRE_CONTROL_WORD, 251, 4, pseudoWireControlWord)
IPFIX_ENTITY(INGRESS_PHYSICAL_INTERFACE, 252, 4, ingressPhysicalInterface)
IPFIX_ENTITY(EGRESS_PHYSICAL_INTERFACE, 253, 4, egressPhysicalInterface)
IPFIX_ENTITY(POST_DOT1Q_VLAN_ID, 254, 2, postDot1qVlanId)
IPFIX_ENTITY(POST_DOT1Q_CUSTOMER_VLAN_ID, 255, 2, postDot1qCustomerVlanId)
IPFIX_ENTITY(ETHERNET_TYPE, 256, 2, ethernetType)
IPFIX_ENTITY(POST_IP_PRECEDENCE, 257, 1, postIpPrecedence)
IPFIX_ENTITY(COLLECTION_TIME_MILLISECONDS, 258, 8, collectionTimeMilliseconds)
IPFIX_ENTITY(EXPORT_SCTP_STREAM_ID, 259, 2, exportSctpStreamId)
IPFIX_ENTITY(MAX_EXPORT_SECONDS, 260, 4, maxExportSeconds)
IPFIX_ENTITY(MAX_FLOW_END_SECONDS, 261, 4, maxFlowEndSeconds)
IPFIX_ENTITY(MESSAGE_MD5_CHECKSUM, 262, 0, messageMD5Checksum)
IPFIX_ENTITY(MESSAGE_SCOPE, 263, 1, messageScope)
IPFIX_ENTITY(MIN_EXPORT_SECONDS, 264, 4, minExportSeconds)
IPFIX_ENTITY(MIN_FLOW_START_SECONDS, 265, 4, minFlowStartSeconds)
IPFIX_ENTITY(OPAQUE_OCTETS, 266, 0, opaqueOctets)
IPFIX_ENTITY(SESSION_SCOPE, 267, 1, sessionScope)
IPFIX_ENTITY(MAX_FLOW_END_MICROSECONDS, 268, 8, maxFlowEndMicroseconds)
IPFIX_ENTITY(MAX_FLOW_END_MILLISECONDS, 269, 8, maxFlowEndMilliseconds)
IPFIX_ENTITY(MAX_FLOW_END_NANOSECONDS, 270, 8, maxFlowEndNanoseconds)
IPFIX_ENTITY(MIN_FLOW_START_MICROSECONDS, 271, 8, minFlowStartMicroseconds)
IPFIX_ENTITY(MIN_FLOW_START_MILLISECONDS, 272, 8, minFlowStartMilliseconds)
IPFIX_ENTITY(MIN_FLOW_START_NANOSECONDS, 273, 8, minFlowStartNanoseconds)
IPFIX_ENTITY(COLLECTOR_CERTIFICATE, 274, 0, collectorCertificate)
IPFIX_ENTITY(EXPORTER_CERTIFICATE, 275, 0, exporterCertificate)
IPFIX_ENTITY(DATA_RECORDS_RELIABILITY, 276, 1, dataRecordsReliability)
IPFIX_ENTITY(OBSERVATION_POINT_TYPE, 277, 1, observationPointType)
IPFIX_ENTITY(CONNECTION_COUNT_NEW, 278, 4, connectionCountNew)
IPFIX_ENTITY(CONNECTION_SUM_DURATION, 279, 8, connectionSumDuration)
IPFIX_ENTITY(CONNECTION_TRANSACTION_ID, 280, 8, connectionTransactionId)
IPFIX_ENTITY(POST_NATSOURCE_IPV6_ADDRESS, 281, 16, postNATSourceIPv6Address)
IPFIX_ENTITY(POST_NATDESTINATION_IPV6_ADDRESS, 282, 16, postNATDestinationIPv6Address)
IPFIX_ENTITY(NAT_POOL_ID, 283, 4, natPoolId)
IPFIX_ENTITY(NAT_POOL_NAME, 284, 0, natPoolName)
IPFIX_ENTITY(ANONYMIZATION_FLAGS, 285, 2, anonymizationFlags)
IPFIX_ENTITY(ANONYMIZATION_TECHNIQUE, 286, 2, anonymizationTechnique)
IPFIX_ENTITY(INFORMATION_ELEMENT_INDEX, 287, 2, informationElementIndex)
IPFIX_ENTITY(P2P_TECHNOLOGY, 288, 0, p2pTechnology)
IPFIX_ENTITY(TUNNEL_TECHNOLOGY, 289, 0, tunnelTechnology)
IPFIX_ENTITY(ENCRYPTED_TECHNOLOGY, 290, 0, encryptedTechnology)
IPFIX_ENTITY(BASIC_LIST, 291, 0, basicList)
IPFIX_ENTITY(SUB_TEMPLATE_LIST, 292, 0, subTemplateList)
IPFIX_ENTITY(SUB_TEMPLATE_MULTI_LIST, 293, 0, subTemplateMultiList)
IPFIX_ENTITY(BGP_VALIDITY_STATE, 294, 1, bgpValidityState)
IPFIX_ENTITY(I_PSEC_SPI, 295, 4, IPSecSPI)
IPFIX_ENTITY(GRE_KEY, 296, 4, greKey)
IPFIX_ENTITY(NAT_TYPE, 297, 1, natType)
IPFIX_ENTITY(INITIATOR_PACKETS, 298, 8, initiatorPackets)
IPFIX_ENTITY(RESPONDER_PACKETS, 299, 8, responderPackets)
IPFIX_ENTITY(OBSERVATION_DOMAIN_NAME, 300, 0, observationDomainName)
IPFIX_ENTITY(SELECTION_SEQUENCE_ID, 301, 8, selectionSequenceId)
IPFIX_ENTITY(SELECTOR_ID, 302, 8, selectorId)
IPFIX_ENTITY(INFORMATION_ELEMENT_ID, 303, 2, informationElementId)
IPFIX_ENTITY(SELECTOR_ALGORITHM, 304, 2, selectorAlgorithm)
IPFIX_ENTITY(SAMPLING_PACKET_INTERVAL, 305, 4, samplingPacketInterval)
IPFIX_ENTITY(SAMPLING_PACKET_SPACE, 306, 4, samplingPacketSpace)
IPFIX_ENTITY(SAMPLING_TIME_INTERVAL, 307, 4, samplingTimeInterval)
IPFIX_ENTITY(SAMPLING_TIME_SPACE, 308, 4, samplingTimeSpace)
IPFIX_ENTITY(SAMPLING_SIZE, 309, 4, samplingSize)
IPFIX_ENTITY(SAMPLING_POPULATION, 310, 4, samplingPopulation)
IPFIX_ENTITY(SAMPLING_PROBABILITY, 311, 8, samplingProbability)
IPFIX_ENTITY(DATA_LINK_FRAME_SIZE, 312, 2, dataLinkFrameSize)
IPFIX_ENTITY(IP_HEADER_PACKET_SECTION, 313, 0, ipHeaderPacketSection)
IPFIX_ENTITY(IP_PAYLOAD_PACKET_SECTION, 314, 0, ipPayloadPacketSection)
IPFIX_ENTITY(DATA_LINK_FRAME_SECTION, 315, 0, dataLinkFrameSection)
IPFIX_ENTITY(MPLS_LABEL_STACK_SECTION, 316, 0, mplsLabelStackSection)
IPFIX_ENTITY(MPLS_PAYLOAD_PACKET_SECTION, 317, 0, mplsPayloadPacketSection)
IPFIX_ENTITY(SELECTOR_ID_TOTAL_PKTS_OBSERVED, 318, 8, selectorIdTotalPktsObserved)
IPFIX_ENTITY(SELECTOR_ID_TOTAL_PKTS_SELECTED, 319, 8, selectorIdTotalPktsSelected)
IPFIX_ENTITY(ABSOLUTE_ERROR, 320, 8, absoluteError)
IPFIX_ENTITY(RELATIVE_ERROR, 321, 8, relativeError)
IPFIX_ENTITY(OBSERVATION_TIME_SECONDS, 322, 4, observationTimeSeconds)
IPFIX_ENTITY(OBSERVATION_TIME_MILLISECONDS, 323, 8, observationTimeMilliseconds)
IPFIX_ENTITY(OBSERVATION_TIME_MICROSECONDS, 324, 8, observationTimeMicroseconds)
IPFIX_ENTITY(OBSERVATION_TIME_NANOSECONDS, 325, 8, observationTimeNanoseconds)
IPFIX_ENTITY(DIGEST_HASH_VALUE, 326, 8, digestHashValue)
IPFIX_ENTITY(HASH_IPPAYLOAD_OFFSET, 327, 8, hashIPPayloadOffset)
IPFIX_ENTITY(HASH_IPPAYLOAD_SIZE, 328, 8, hashIPPayloadSize)
IPFIX_ENTITY(HASH_OUTPUT_RANGE_MIN, 329, 8, hashOutputRangeMin)
IPFIX_ENTITY(HASH_OUTPUT_RANGE_MAX, 330, 8, hashOutputRangeMax)
IPFIX_ENTITY(HASH_SELECTED_RANGE_MIN, 331, 8, hashSelectedRangeMin)
IPFIX_ENTITY(HASH_SELECTED_RANGE_MAX, 332, 8, hashSelectedRangeMax)
IPFIX_ENTITY(HASH_DIGEST_OUTPUT, 333, 1, hashDigestOutput)
IPFIX_ENTITY(HASH_INITIALISER_VALUE, 334, 8, hashInitialiserValue)
IPFIX_ENTITY(SELECTOR_NAME, 335, 0, selectorName)
IPFIX_ENTITY(UPPER_CILIMIT, 336, 8, upperCILimit)
IPFIX_ENTITY(LOWER_CILIMIT, 337, 8, lowerCILimit)
IPFIX_ENTITY(CONFIDENCE_LEVEL, 338, 8, confidenceLevel)
IPFIX_ENTITY(INFORMATION_ELEMENT_DATA_TYPE, 339, 1, informationElementDataType)
IPFIX_ENTITY(INFORMATION_ELEMENT_DESCRIPTION, 340, 0, informationElementDescription)
IPFIX_ENTITY(INFORMATION_ELEMENT_NAME, 341, 0, informationElementName)
IPFIX_ENTITY(INFORMATION_ELEMENT_RANGE_BEGIN, 342, 8, informationElementRangeBegin)
IPFIX_ENTITY(INFORMATION_ELEMENT_RANGE_END, 343, 8, informationElementRangeEnd)
IPFIX_ENTITY(INFORMATION_ELEMENT_SEMANTICS, 344, 1, informationElementSemantics)
IPFIX_ENTITY(INFORMATION_ELEMENT_UNITS, 345, 2, informationElementUnits)
IPFIX_ENTITY(PRIVATE_ENTERPRISE_NUMBER, 346, 4, privateEnterpriseNumber)
IPFIX_ENTITY(VIRTUAL_STATION_INTERFACE_ID, 347, 0, virtualStationInterfaceId)
IPFIX_ENTITY(VIRTUAL_STATION_INTERFACE_NAME, 348, 0, virtualStationInterfaceName)
IPFIX_ENTITY(VIRTUAL_STATION_UUID, 349, 0, virtualStationUUID)
IPFIX_ENTITY(VIRTUAL_STATION_NAME, 350, 0, virtualStationName)
IPFIX_ENTITY(LAYER2_SEGMENT_ID, 351, 8, layer2SegmentId)
IPFIX_ENTITY(LAYER2_OCTET_DELTA_COUNT, 352, 8, layer2OctetDeltaCount)
IPFIX_ENTITY(LAYER2_OCTET_TOTAL_COUNT, 353, 8, layer2OctetTotalCount)
IPFIX_ENTITY(INGRESS_UNICAST_PACKET_TOTAL_COUNT, 354, 8, ingressUnicastPacketTotalCount)
IPFIX_ENTITY(INGRESS_MULTICAST_PACKET_TOTAL_COUNT, 355, 8, ingressMulticastPacketTotalCount)
IPFIX_ENTITY(INGRESS_BROADCAST_PACKET_TOTAL_COUNT, 356, 8, ingressBroadcastPacketTotalCount)
IPFIX_ENTITY(EGRESS_UNICAST_PACKET_TOTAL_COUNT, 357, 8, egressUnicastPacketTotalCount)
IPFIX_ENTITY(EGRESS_BROADCAST_PACKET_TOTAL_COUNT, 358, 8, egressBroadcastPacketTotalCount)
IPFIX_ENTITY(MONITORING_INTERVAL_START_MILLI_SECONDS, 359, 8, monitoringIntervalStartMilliSeconds)
IPFIX_ENTITY(MONITORING_INTERVAL_END_MILLI_SECONDS, 360, 8, monitoringIntervalEndMilliSeconds)
IPFIX_ENTITY(PORT_RANGE_START, 361, 2, portRangeStart)
IPFIX_ENTITY(PORT_RANGE_END, 362, 2, portRangeEnd)
IPFIX_ENTITY(PORT_RANGE_STEP_SIZE, 363, 2, portRangeStepSize)
IPFIX_ENTITY(PORT_RANGE_NUM_PORTS, 364, 2, portRangeNumPorts)
IPFIX_ENTITY(STA_MAC_ADDRESS, 365, 6, staMacAddress)
IPFIX_ENTITY(STA_IPV4_ADDRESS, 366, 4, staIPv4Address)
IPFIX_ENTITY(WTP_MAC_ADDRESS, 367, 6, wtpMacAddress)
IPFIX_ENTITY(INGRESS_INTERFACE_TYPE, 368, 4, ingressInterfaceType)
IPFIX_ENTITY(EGRESS_INTERFACE_TYPE, 369, 4, egressInterfaceType)
IPFIX_ENTITY(RTP_SEQUENCE_NUMBER, 370, 2, rtpSequenceNumber)
IPFIX_ENTITY(USER_NAME, 371, 0, userName)
IPFIX_ENTITY(APPLICATION_CATEGORY_NAME, 372, 0, applicationCategoryName)
IPFIX_ENTITY(APPLICATION_SUB_CATEGORY_NAME, 373, 0, applicationSubCategoryName)
IPFIX_ENTITY(APPLICATION_GROUP_NAME, 374, 0, applicationGroupName)
IPFIX_ENTITY(ORIGINAL_FLOWS_PRESENT, 375, 8, originalFlowsPresent)
IPFIX_ENTITY(ORIGINAL_FLOWS_INITIATED, 376, 8, originalFlowsInitiated)
IPFIX_ENTITY(ORIGINAL_FLOWS_COMPLETED, 377, 8, originalFlowsCompleted)
IPFIX_ENTITY(DISTINCT_COUNT_OF_SOURCE_IPADDRESS, 378, 8, distinctCountOfSourceIPAddress)
IPFIX_ENTITY(DISTINCT_COUNT_OF_DESTINATION_IPADDRESS, 379, 8, distinctCountOfDestinationIPAddress)
IPFIX_ENTITY(DISTINCT_COUNT_OF_SOURCE_IPV4_ADDRESS, 380, 4, distinctCountOfSourceIPv4Address)
IPFIX_ENTITY(DISTINCT_COUNT_OF_DESTINATION_IPV4_ADDRESS, 381, 4, distinctCountOfDestinationIPv4Address)
IPFIX_ENTITY(DISTINCT_COUNT_OF_SOURCE_IPV6_ADDRESS, 382, 8, distinctCountOfSourceIPv6Address)
IPFIX_ENTITY(DISTINCT_COUNT_OF_DESTINATION_IPV6_ADDRESS, 383, 8, distinctCountOfDestinationIPv6Address)
IPFIX_ENTITY(VALUE_DISTRIBUTION_METHOD, 384, 1, valueDistributionMethod)
IPFIX_ENTITY(RFC3550_JITTER_MEAN_MILLISECONDS, 385, 4, rfc3550JitterMeanMilliseconds)
IPFIX_ENTITY(RFC3550_JITTER_MEAN_MICROSECONDS, 386, 4, rfc3550JitterMeanMicroseconds)
IPFIX_ENTITY(RFC3550_JITTER_MEAN_NANOSECONDS, 387, 4, rfc3550JitterMeanNanoseconds)

#undef IPFIX_ENTITY
